# coding: utf-8

"""Some functions to query the Bitbucket repository"""

import os
import json
import urllib2

from aslint.logger import logger
from aslint.i18n import _
from aslint.baseutils import force_list, NeverFailedDict


BBURL = 'https://bitbucket.org/code_aster/codeaster-src/issue/%s'

LOG_TEMPL = """
--------------------------------------------------------------------------------
--- List of Bitbucket issues integrated in this release
"""

LOG_ISSUE = """
------------------------------------------------------------------------
Title
    %(title)s
Assigned to: %(resp)s
URL: %(url)s

Comments:
%(comments)s
"""

LOG_CMT = """%(author)s - %(time)s:
%(comment)s

"""


def _ident(txt, level):
    """Ident the text of level*N spaces"""
    ident = ' ' * 4 * level
    return os.linesep.join([ident + line for line in txt.splitlines()])


def get_bitbucket_url(issue):
    """Return the url of a bitbucket issue"""
    return BBURL % issue


def request_api(issue, method):
    """Request for a API method"""
    url = 'https://bitbucket.org/api/1.0/repositories/' \
        'code_aster/codeaster-src/issues/%(issue)s/%(meth)s' \
        % {'issue': issue, 'meth': method}
    data = {}
    try:
        url = urllib2.urlopen(url, timeout=3)
        data = json.loads(url.read())
    except urllib2.URLError, exc:
        logger.error(_("connection to bitbucket failed (url: %(url)s):\n"
                       "%(exc)s"), {'url': url, 'exc': str(exc)})
    return data


def get_bitbucket_issue(issue):
    """Return the title of Bitbucket issue"""
    issue_content = NeverFailedDict({
        'title': 'bb%04d' % issue,
        'comments': '',
        'url': get_bitbucket_url(issue)})
    data = NeverFailedDict(request_api(issue, ''))
    if not data:
        return issue_content
    issue_content['title'] = '[bb%04d] - %s' % (issue, data['title'])
    issue_content['resp'] = data['responsible']['display_name']
    comments = force_list(request_api(issue, 'comments/'))
    if not comments:
        return issue_content
    txt = []
    # creator
    txt.append(LOG_CMT % {'author': data['reported_by']['display_name'],
                          'time': data['utc_created_on'],
                          'comment': _ident(data['content'] or '', 1)})
    comments.reverse()
    for cmt in comments:
        cmt = NeverFailedDict(cmt)
        if cmt['content']:
            author = cmt['author_info']['display_name'] \
                if cmt['author_info'] else "anonymous"
            txt.append(LOG_CMT % {
                'author': author,
                'time': cmt['utc_created_on'],
                'comment': _ident(cmt['content'] or '', 1)})
    issue_content['comments'] = os.linesep.join([_ident(i, 1) for i in txt])
    return issue_content


def get_bitbucket_log(issues):
    """Return a text for the changelog of issues solved on Bitbucket"""
    text = []
    if issues:
        text.append(LOG_TEMPL)
    for num in issues:
        logger.debug(_("getting Bitbucket issue #%s"), num)
        text.append(LOG_ISSUE % get_bitbucket_issue(num))
    return os.linesep.join(text)
