# coding: utf-8

"""Interfaces to call asrun functions"""

import sys
import os
import os.path as osp
from functools import partial
import tempfile
from functools import wraps
import subprocess as SP

from aslint.i18n import _
from aslint.logger import logger
from aslint.config import ASCFG


def check_asrun():
    """Check import of asrun"""
    check_asrun_bin()
    asrun_lib = ASCFG.get('asrun.lib')
    if not asrun_lib in sys.path:
        sys.path.append(asrun_lib)
    try:
        import asrun.run
        assert hasattr(asrun.run, 'AsRunFactory')
    except (ImportError, AssertionError):
        logger.error(_("please check the parameter 'asrun.bin' is correct in "
                       "your ~/.hgrc configuration file, if it is, you may "
                       "have to set manually 'asrun.lib'"))
        raise


def check_asrun_bin():
    """Check availability of as_run script"""
    try:
        SP.check_call([ASCFG.get('asrun.bin'), '--version'], stdout=SP.PIPE)
    except (OSError, SP.CalledProcessError):
        logger.error(_("'as_run' script not found in your $PATH; sourcing an "
                       "environment may be required. You can also "
                       "define 'asrun.bin' in your ~/.hgrc configuration file "
                       "to select a specific version."))
        raise


def need_asrun(func):
    """decorator for functions that require asrun package"""
    @wraps(func)
    def wrapper(*args, **kwds):
        """wrapper"""
        check_asrun()
        return func(*args, **kwds)
    return wrapper


@need_asrun
def get_asrun_config(version_path):
    """Return the :py:class:`AsterConfig` object for a version.

    Arguments:
        version_path (str): Path to the version.

    Returns:
        *AsterConfig*: asrun object.
    """
    from asrun.config import AsterConfig
    config = AsterConfig(osp.join(version_path, 'config.txt'))
    return config


REX_ERR = _("error during connection to REX database")


@need_asrun
def _get_rex_connection(write=False):
    """Return a CONNECT object to the REX database"""
    try:
        import asrun.myconnect as DB
        from asrun.schema import MySQLError
    except ImportError:
        logger.warn(_("perhaps MySQLdb is missing or the environment is not "
                      "properly set, try:\n") +
                    "    . $HOME/dev/codeaster/devtools/etc/env_unstable.sh\n"
                    " or\n"
                    "    . $HOME/dev/codeaster/devtools/etc/env_stable-updates.sh\n")
        logger.warn(REX_ERR)
        return None
    kwargs = {}
    if not write:
        kwargs['db_params'] = ASCFG.get('rex_params')
    try:
        cnx = DB.CONNECT('REX', verbose=logger.debug_enabled(), **kwargs)
    except MySQLError:
        logger.warn(_("asrun 1.13.1 or newer is required"))
        logger.warn(REX_ERR)
        return None
    return cnx


@need_asrun
def get_load_issue():
    """Return a function to load an issue: func(#issue)"""
    cnx = _get_rex_connection()
    if not cnx:
        return None
    # here because import is checked in _get_rex_connection
    from asrun.schema import ISSUE
    return partial(ISSUE, c=cnx)



@need_asrun
def get_histor(issue_list, fmt='text'):
    """Build a text of the messages of the given issues"""
    from asrun.run import AsRunFactory
    from asrun.rex import build_histor
    run = AsRunFactory(debug_stderr=False)
    run.options['format'] = fmt
    run.options['status'] = 'all'
    cnx = _get_rex_connection()
    if not cnx:
        logger.error(REX_ERR)
        return REX_ERR
    histor = build_histor(run, issue_list, cnx)
    return repr(histor)


@need_asrun
def get_rex_users(fields, crit=None):
    """Return the list of users, each dict containing the `fields`"""
    assert type(fields) == list, '`fields` argument must be a list'
    cnx = _get_rex_connection()
    if not cnx:
        return None
    from asrun.schema import ReadDB, USER
    all_usr = ReadDB(USER, cnx)
    fields.append('__retired__')
    not_retired = "__retired__ == 0"
    if crit is None:
        crit = not_retired
    else:
        crit = "%s and (%s)" % (not_retired, crit)
    users = {}
    for username, values in all_usr.items():
        new = {}
        for key in fields:
            val = values.get(key)
            if type(val) in (str, unicode, int, long):
                new[key] = val
        if new:
            new['_username'] = username
            try:
                match = eval(crit, new.copy())
            except NameError:
                match = None
            if match:
                users[username] = new
    return users
