# coding=utf-8

"""Function to check the installation"""

import os
import os.path as osp
from glob import glob
import tempfile
import shutil
from subprocess import Popen, PIPE, STDOUT

from aslint.logger import logger, ErrorCode
from aslint.i18n import _


def check_install(c4che, dry_run=False):
    """Check installation"""
    asterxx = c4che.get('BOOST_VERSION') is not None
    try:
        if c4che.get('ASTERBINOPT'):
            # new - see BB0006
            executable = osp.join(c4che['BINDIR'], c4che['ASTERBINOPT'])
            bibpyt = c4che['ASTERLIBDIR']
        else:
            executable = osp.join(c4che['BINDIR'], 'aster')
            bibpyt = c4che['PYTHONARCHDIR']
        python = c4che['PYTHON'][0]
        prefix = c4che['PREFIX']
        datadir = c4che['ASTERDATADIR']
    except (TypeError, KeyError):
        logger.error(_("waf parameters are required"))
        return logger.errcode
    logger.title(_("checking installation in %s"), prefix)
    errcode = ErrorCode()
    expected = []
    if not asterxx:
        expected.append(executable)
    expected.extend([osp.join(datadir, path)
                     for path in ('config.txt', 'profile.sh')])
    expected.extend([osp.join(bibpyt, path)
                     for path in ('code_aster', 'Execution', 'Validation')])

    for path in expected:
        if len(glob(path)) != 1:
            msg = _('not found: {0}').format(path)
            if not dry_run:
                logger.error(msg)
            else:
                logger.info(msg)
    cata = [osp.join(bibpyt, path)
            for path in (osp.join('code_aster', 'Cata'), 'Cata')]
    lcata = [len(glob(path)) for path in cata]
    if sum(lcata) < 1:
        # currently Cata still exists even if code_aster/Cata exists (i18n files)
        msg = _('at least one of {0} is expected').format(tuple(cata))
        if not dry_run:
            logger.error(msg)
        else:
            logger.info(msg)
    errcode.add(check_import_cata(python, bibpyt, dry_run=dry_run))
    if errcode.is_error():
        if not dry_run:
            logger.error(_("invalid installation"))
        else:
            logger.info(_("(DRY RUN) invalid installation"))
    else:
        logger.info(_("installation ok"))
    return errcode


def check_import_cata(python, bibpyt, dry_run=False):
    """Check the cata.py can be imported with only the minimal number
    of required packages"""
    errcode = ErrorCode()

    def _exec(cmd):
        """Execute an external command"""
        if not dry_run:
            proc = Popen(cmd, stdout=PIPE, stderr=STDOUT)
            errcode.from_status(proc.wait())
        else:
            logger.info(_("(DRY RUN) %s"), cmd)
    dtmp = tempfile.mkdtemp()
    if osp.exists(osp.join(bibpyt, 'code_aster', 'Cata')):
        supv = (osp.join('code_aster', 'Cata'), )
    else:
        supv = ('Accas', 'Build', 'Cata', 'Execution', 'Noyau', 'Validation')
    prev = os.getcwd()
    os.chdir(dtmp)
    try:
        if not dry_run:
            for pkg in supv:
                shutil.copytree(osp.join(bibpyt, pkg), pkg)
            open('aster.py', 'w').write(os.linesep.join([
                "class error(Exception):",
                "   pass",
                "class FatalError(Exception):",
                "   pass",
                "def affiche(unit, txt):",
                "   print txt",
                "def onFatalError():",
                "   return 'ABORT'",
            ]))
        _exec([python, '-c', "'import Cata'"])
        _exec([python, '-c', "'from Cata import cata'"])
    finally:
        os.chdir(prev)
        shutil.rmtree(dtmp)
    return errcode
