# coding=utf-8

"""Some string utilities"""

import re

from aslint.baseutils import get_encoding


def to_unicode(string):
    """Try to convert string into a unicode string
    """
    if type(string) is unicode:
        return string
    for encoding in ('utf-8', 'iso-8859-15', 'cp1252'):
        try:
            s = unicode(string, encoding)
            return s
        except UnicodeDecodeError:
            pass
    return unicode(string, 'utf-8', 'replace')


def from_unicode(ustring, encoding, errors='replace'):
    """Try to encode a unicode string using encoding.
    """
    try:
        return ustring.encode(encoding)
    except UnicodeError:
        pass
    return ustring.encode(encoding, errors)


def convert(content, encoding=None, errors='replace'):
    """Convert content using encoding or default encoding if None."""
    if type(content) not in (str, unicode):
        content = unicode(content)
    if type(content) == str:
        content = to_unicode(content)
    return from_unicode(content, encoding or get_encoding(), errors)


# Levenshtein algorithm implementation
# original author: Magnus Lie Hetland
# License <http://creativecommons.org/publicdomain/zero/1.0>
def levenshtein(a, b):
    """Calculate the Levenshtein distance between a and b."""
    n, m = len(a), len(b)
    if n > m:
        # Make sure n <= m, to use O(min(n,m)) space
        a, b = b, a
        n, m = m, n
    current = range(n + 1)
    for i in range(1, m + 1):
        previous, current = current, [i] + [0] * n
        for j in range(1, n + 1):
            add, delete = previous[j] + 1, current[j - 1] + 1
            change = previous[j - 1]
            if a[j - 1] != b[i - 1]:
                change = change + 1
            current[j] = min(add, delete, change)
    return current[n]


def relative_levenshtein(a, b):
    """Calculate a relative Levenshtein distance between a and b."""
    a = remove_accent(a)
    b = remove_accent(b)
    norm = max((len(a) + len(b)) / 2., 1.)
    return levenshtein(a, b) / norm


def remove_accent(text):
    """Remove accentuated characters"""
    lreg = []
    text = to_unicode(text)
    for pattern, repl in ([u'[àâä]', 'a'], [u'[éèêë]', 'e'], [u'[îï]', 'i'],
                          [u'[ôö]', 'o'], [u'[ùûü]', 'u'],
                          [u'[ÀÂÄ]', 'A'], [u'[ÉÈÊË]', 'E'], [u'[ÎÏ]', 'I'],
                          [u'[ÔÖ]', 'O'], [u'[ÙÛÜ]', 'U']):
        text = re.sub(pattern, repl, text)
    return convert(text)


def search_similar(text, values, maxdist=1):
    """Search similar text in values, return the nearest value with relative Levenshtein distance less than 'maxdist'."""
    assert len(values) > 0
    scores = sorted([(relative_levenshtein(text, other), other) for other in values])
    score, found = scores[0]
    if score <= maxdist:
        return found
    return None
