# coding=utf-8

"""Checkers for testcase files"""

import os
import os.path as osp
import stat
import re

from aslint.api_asrun import check_asrun
from aslint.utils import read_testlist, checksum
from aslint.base_checkers import (
    TextMsg, TextMsgNotFound, GenericMsg, checkers_from_context, search_msg,
    DirnameCat, FilenameCat, FileContentCat, DiffCat, CheckContext,
)
import aslint.common_checkers as COMM
from aslint.config import ASCFG

COMM_EXT = ['.com' + i for i in 'm0123456789']
REG_EXT = ("com[m0-9]", "py", "data", "datg", "export", "[0-9]{1,2}",
           "mail", "mgib", "med", "mmed", "msh", "msup", "mfront")


class UnsupportedExtension(FilenameCat, TextMsg):

    """Unsupported extension
    Only these extensions are allowed for testcases: .comm, .com[0-9], .py,
    .data, .datg, .export, .mail, .mgib, .mmed, .msh, .msup, .[0-9]+"""
    id = "C9002"
    fmt = "%(id)s: %(label)s"
    _search = search_msg(r"(?P<ext>\.(" + "|".join(REG_EXT) + "))$")

    def search(self, fname):
        """Check for unsupported extension."""
        if not osp.isfile(fname) or self._search(fname):
            return []
        return [fname]


class InvalidFilename(FilenameCat, TextMsgNotFound):

    """Invalid filename for a testcase
    Filenames matching 'aaaaNNNm.*' are expected.
    See C9002 message for the allowed extensions.
    """
    id = "C2015"
    fmt = "%(id)s: %(label)s"
    apply_ctxt = CheckContext(reponame=('src', 'validation'),
                              branch='^(default|v[0-9]+)')

    def search(self, txt):
        """Check for unauthorized filenames."""
        # most of names are: aaaaNNNm
        # - some have only 2 numbers,
        # - aspic/ascouf need 5 letters,
        # - mac3c needs the '3'
        _expr = re.compile(r"/(?P<root>[a-z3]{3,5}[0-9]{2,3}[a-z]{1})"
                           r"(?P<ext>\.(" + "|".join(REG_EXT) + "))$")
        found = _expr.findall(txt)
        # accept .mfront files too
        if not found and osp.splitext(txt)[1] == '.mfront':
            found = [txt]
        return found


class InvalidFilenameTestsData(FilenameCat, TextMsgNotFound):

    """Invalid filename for a testcase
    Filenames matching 'aaaaNNNm.*' or 'aaaaNNNm_*.*' are expected.
    See C9002 message for the allowed extensions.
    """
    id = "C2019"
    fmt = "%(id)s: %(label)s"
    apply_ctxt = CheckContext(reponame=('data', ),
                              branch='^(default|v[0-9]+)')

    def search(self, txt):
        """Check for unauthorized filenames."""
        # most of names are: aaaaNNNm
        # - some have only 2 numbers,
        # - aspic/ascouf need 5 letters,
        # - mac3c needs the '3'
        _expr1 = re.compile(r"/(?P<root>[a-z3]{3,5}[0-9]{2,3}[a-z]{1})"
                            r"(?P<ext>\.(" + "|".join(REG_EXT) + "))$")
        _expr2 = re.compile(r"/(?P<root>[a-z3]{3,5}[0-9]{2,3}[a-z]{1}"
                            r"_\w{1,16})"
                            r"(?P<ext>\.(" + "|".join(REG_EXT) + "))$")
        found = (_expr1.findall(txt) or _expr2.findall(txt))
        # accept .mfront files too
        if not found and osp.splitext(txt)[1] == '.mfront':
            found = [txt]
        return found


class InvalidFilenameXX(FilenameCat, TextMsgNotFound):

    """Invalid filename for a testcase
    Filenames matching 'aaaaNNNm.*', 'aaaaNNNm_*.comm', 'aaaaNNNm_*.py',
    'xx*.export', 'xx*.py' or 'NNN.med' are expected.
    See C9002 message for the allowed extensions.
    """
    id = "C2016"
    fmt = "%(id)s: %(label)s"
    apply_ctxt = CheckContext(branch='^asterxx')

    def search(self, txt):
        """Check for unauthorized filenames."""
        # most of names are: aaaaNNNm
        # - some have only 2 numbers,
        # - aspic/ascouf need 5 letters,
        # - mac3c needs the '3'
        # accept aaaaNNNm + a suffix starting with '_'
        # accept 'xx*'
        _expr1 = re.compile(r"/(?P<root>[a-z3]{3,16}[0-9]{2,3}[a-z]{1})"
                            r"(?P<ext>\.(" + "|".join(REG_EXT) + "))$")
        _expr2 = re.compile(r"/(?P<root>[a-z3]{3,16}[0-9]{2,3}[a-z]{1}"
                            r"_\w{1,16})"
                            r"(?P<ext>\.(comm|py))$")
        _expr3 = re.compile(r"/(?P<root>xx\w+)"
                            r"(?P<ext>(|\.(" + "|".join(REG_EXT) + ")))$")
        _expr4 = re.compile(r"/(?P<root>[0-9]+)"
                            r"(?P<ext>\.med)$")
        found = (_expr1.findall(txt) or
                 _expr2.findall(txt) or
                 _expr3.findall(txt) or
                 _expr4.findall(txt))
        # accept .mfront files too
        if not found and osp.splitext(txt)[1] == '.mfront':
            found = [txt]
        return found


class FileNotUsed(DirnameCat, GenericMsg):

    """File not used by any testcase
    A file is in the directory of the testcases but it is not used by any
    testcase."""
    id = "C2004"
    re_expr = re.compile('^[FR] +[^ ]+ +([^ ]+) +D', re.M)

    def search(self, txt):
        """Check if files exist"""
        # creating an AsterProfil object for each export costs twice more
        astest = osp.join(txt, 'astest')
        files = set(os.listdir(astest))
        files.discard('wscript')
        lexp = [fname for fname in files if fname.endswith('.export')]
        files.difference_update(lexp)
        ltxt = []
        for path in lexp:
            ltxt.append(open(osp.join(astest, path), 'rb').read())
        text = os.linesep.join(ltxt)
        used = self.re_expr.findall(text)
        unused = [': %s' % path for path in files.difference(used)]
        return unused


class DatafileNotFound(DirnameCat, GenericMsg):

    """Data file not found
    The datafile referenced in the export file does not exist."""
    id = 'E2001'
    re_expr = re.compile('^[FR] +[^ ]+ +([^ ]+) +D', re.M)

    def search(self, txt):
        """Check if files exist"""
        # creating an AsterProfil object for each export costs twice more
        astest = osp.join(txt, 'astest')
        tests_data = osp.join(txt, '..', 'data', 'tests_data')
        files = set(os.listdir(astest))
        lexp = [fname for fname in files if fname.endswith('.export')]
        ltxt = []
        for path in lexp:
            ltxt.append(open(osp.join(astest, path), 'rb').read())
        text = os.linesep.join(ltxt)
        used = self.re_expr.findall(text)
        notfound = []
        for path in used:
            if (not osp.exists(osp.join(astest, path)) and
                    not osp.exists(osp.join(tests_data, path))):
                notfound.append(': %s' % path)
        return notfound


class TestInSubmitList(DirnameCat, GenericMsg):

    """Testcase in the submit list removed
    A testcase is in the list of the testcases passed during the submit but this
    testcase does not exist anymore. Please contact the administrator to remove
    the testcase from the list."""
    id = 'W2010'
    apply_ctxt = CheckContext(branch=['default'], reponame=['src'])

    def search(self, txt):
        """Check if files exist"""
        astest = osp.join(txt, 'astest')
        files = set(os.listdir(astest))
        lexp = [fname for fname in files if fname.endswith('.export')]
        ltst = [osp.splitext(osp.basename(fname))[0] for fname in lexp]
        flist = osp.join(ASCFG.get('aslint_root'), 'test', 'list.submit')
        tlist = read_testlist(flist)
        notfound = [': %s' % path for path in tlist.difference(ltst)]
        return notfound


class DatafileTooBig(FilenameCat, GenericMsg):

    """Size of datafiles too big"""
    apply_to = ['.export']
    id = 'C2002'
    limit = 2. * 1048576
    apply_ctxt = CheckContext(reponame=['src'])

    def search(self, txt):
        """Check size of files"""
        if not osp.isfile(txt):
            return []
        check_asrun()
        from asrun.profil import AsterProfil
        if osp.basename(txt) in ['perf011a.export',  'perf011b.export']:
            # share a 400 MB file...
            return []
        pexp = AsterProfil(txt)
        dname = osp.dirname(txt)
        results = []
        size = 0
        for entry in pexp.get_data():
            path = osp.join(dname, entry.path)
            if osp.exists(path):
                size += os.stat(path)[stat.ST_SIZE]
        if size > self.limit:
            results.append(': {0:.3f} MB (limit: {1:.3f} MB)'
                           .format(size / 1048576., self.limit / 1048576.))
        return results


class DatafileTooBigV(DatafileTooBig):
    """Size of datafiles too big in validation repository"""
    id = 'C2018'
    limit = 65. * 1048576
    apply_ctxt = CheckContext(reponame=['validation'])


class InvalidExport(GenericMsg):
    """Invalid export file
    Required parameters: memory_limit, time_limit, testlist.

    testlist must contain 'verification', 'verif_interact' or 'validation',
    and, 'sequential' or 'parallel'.

    Additional supported names:
        'benchmark', 'concrete', 'dam', 'fatigue', 'fluid_structure',
        'fracture', 'fuel_assembly', 'performance', 'prototype', 'seism',
        'welding'.
    """
    apply_to = ['.export']
    auth = ('memory_limit', 'time_limit',
            'ncpus', 'mpi_nbnoeud', 'mpi_nbcpu', 'max_base',
            'testlist', 'expected_diag')
    required = ('memory_limit', 'time_limit', 'testlist', )
    testlist_names = ()
    project_list = ()

    def search(self, txt):
        """Check a .export. Return True if it is ok."""
        check_asrun()
        from asrun.profil import AsterProfil
        # check asrun compatibility
        import inspect
        args = inspect.getargspec(AsterProfil.__init__).args
        opts = {}
        if 'auto_param' in args:
            opts['auto_param'] = False
        pexp = AsterProfil(txt, **opts)
        results = []
        for key in pexp.param.keys():
            if not key in self.auth:
                results.append(": unauthorized parameter: %s" % key)
        for key in self.required:
            if pexp.param.get(key) is None:
                results.append(": required parameter: %s" % key)
        testlist = pexp.param.get('testlist', '')
        if type(testlist) in (list, tuple):
            testlist = ' '.join(testlist)
        verif = 'verification' in testlist
        verif_inter = 'verif_interact' in testlist
        valid = 'validation' in testlist
        seq = 'sequential' in testlist
        par = 'parallel' in testlist
        if (not (verif or verif_inter or valid)) \
            or (verif and valid):
            results.append(": expecting exactly one of 'verification', "
                           "'verif_interact' or 'validation' in testlist")
        if (not (seq or par)) or (seq and par):
            results.append(": expecting exactly one of sequential or "
                           "parallel in testlist")
        # check for unknown list
        for name in testlist.split():
            if name not in list(self.testlist_names) + list(self.project_list):
                results.append(": unexpected list name: %r" % name)
        if valid:
            # project is required
            project_found = False
            for name in self.project_list:
                if name in testlist.split():
                    project_found = True
                    break
            if not project_found:
                results.append(": a valid domain name is required "
                               "(none in 'src'), one of {0}"
                               .format(self.project_list))
        if not valid and pexp.param.get('mpi_nbcpu'):
            try:
                value = int(pexp.param.get('mpi_nbcpu')[0])
                if value > 8:
                    results.append(
                        ": mpi_nbcpu is limited to 8 for verification")
            except (TypeError, ValueError):
                results.append(": integer value is expected for mpi_nbcpu")
        for entry in pexp.get_data():
            orig = entry.path.strip()
            basn = osp.basename(orig)
            if orig != basn:
                results.append(": must be a filename without path: %s"
                               % entry.path)
        result = pexp.get_result()
        if len(result) != 0:
            results.append(": must contain no result: %d found" % len(result))
        larg = len(pexp.args)
        if larg != 0:
            results.append(": must contain no argument: %d found" % larg)
        return results


class InvalidExportVerif(FilenameCat, InvalidExport):
    """Invalid export file
    Required parameters: memory_limit, time_limit, testlist.

    testlist must contain 'verification' or 'verif_interact',
    and, 'sequential' or 'parallel'.
    """
    apply_ctxt = CheckContext(reponame=['src'])
    id = 'E2003'
    testlist_names = ('verification', 'verif_interact',
                      'sequential', 'parallel', 'asterxx')


class InvalidExportValid(FilenameCat, InvalidExport):
    """Invalid export file
    Required parameters: memory_limit, time_limit, testlist.

    testlist must contain 'validation' and, 'sequential' or 'parallel'.

    Additional supported names:
        'benchmark', 'concrete', 'dam', 'fatigue', 'fluid_structure',
        'fracture', 'fuel_assembly', 'performance', 'prototype', 'seism',
        'welding'.
    """
    apply_ctxt = CheckContext(reponame=['validation'])
    id = 'E2103'
    testlist_names = ('validation', 'verif_interact',
                      'sequential', 'parallel', 'asterxx')
    project_list = (
        'benchmark', 'concrete', 'dam', 'fatigue', 'fluid_structure',
        'fracture', 'fuel_assembly', 'performance', 'prototype', 'seism',
        'welding'
    )


class ResourcesTooBig(FilenameCat, GenericMsg):

    """Testcase resources are too big
    Time limit must be less than 300 s, memory limit less than 1024 MB in 'src'.
    This error can be ignored in the 'validation' repository"."""
    apply_to = ['.export']
    id = 'C2009'
    apply_ctxt = CheckContext(reponame=['src'])

    def search(self, txt):
        """Check a .export"""
        if not osp.isfile(txt):
            return []
        check_asrun()
        from asrun.profil import AsterProfil
        pexp = AsterProfil(txt)
        results = []
        try:
            time_limit = int(pexp['time_limit'][0])
            assert time_limit <= 300
        except (AssertionError, ValueError, TypeError):
            val = pexp['time_limit'][0]
            results.append(": invalid parameter, time_limit must be an "
                           "integer less than 300: %s" % val)
        try:
            time_limit = int(pexp['memory_limit'][0])
            assert time_limit <= 1024
        except (AssertionError, ValueError, TypeError):
            val = pexp['memory_limit'][0]
            results.append(": invalid parameter, memory_limit must be an "
                           "integer less than 1024: %s" % val)
        return results


class DuplicatedDataFile(DirnameCat, GenericMsg):

    """Duplicated data files
    Datafiles can be shared between testcases through the .export definition.
    Do not add the same file several times."""
    id = 'C2013'

    def search(self, txt):
        """Check for duplicate datafiles"""
        err = []
        astest = osp.join(txt, 'astest')
        files = set(os.listdir(astest))
        dsize = {}
        for fname in files:
            path = osp.join(astest, fname)
            if not osp.isfile(path):
                continue
            size = os.stat(path)[stat.ST_SIZE]
            # skip small files
            if size < 1000:
                continue
            dsize[size] = dsize.get(size, []) + [path]
        for size, files in dsize.items():
            if len(files) < 2:
                continue
            dble = {}
            for fname in files:
                fsum = checksum(fname)
                dble[fsum] = dble.get(fsum, []) + [fname]
            for files in dble.values():
                if len(files) > 1:
                    files.sort()
                    err.append(": %s" % ', '.join(files))
        return err


class BaseUnrecommendedKeyword(TextMsg):

    """Unrecommended keyword in a command file
    This keyword should not be used in a testcase: TOLE_MACHINE."""
    # id must be defined in subclasses
    kwd = ['TOLE_MACHINE', ]
    search = search_msg("^ *[^\#].*(?P<main>(%s))" % '|'.join(kwd))


class UnrecommendedKeyword(FileContentCat, BaseUnrecommendedKeyword):

    """Unrecommended keyword in a command file
    This keyword should not be used in a testcase: TOLE_MACHINE."""
    id = 'W2005'


class NewUnrecommendedKeyword(DiffCat, BaseUnrecommendedKeyword):

    """Unrecommended keyword in a command file
    This keyword should not be used in a testcase: TOLE_MACHINE."""
    id = 'C2055'


class InvalidKeywordValue(FileContentCat, GenericMsg):

    """Unrecommended value assigned to a keyword in a command file
    This value should be assigned wit care in this repository."""
    id = 'W2012'
    _invalid = [('NIV_PUB_WEB', '[\'\"]INTRANET[\'\"]'), ]
    _expr = [re.compile("^ *[^\#].*(?P<kwd>%s) *= *(?P<value>%s)"
             % (kwd, val), re.M) for kwd, val in _invalid]
    apply_ctxt = CheckContext(reponame=['src'])

    def search(self, txt):
        """Check a command file for invalid values"""
        err = []
        for expr in self._expr:
            mat = expr.search(txt)
            if mat:
                err.append(": %s=%s" % (mat.group('kwd'), mat.group('value')))
        return err


class DeprecatedComment(FileContentCat, TextMsg):

    """Deprecated comment in a command file
    This comment (TITRE) should not be used in a testcase: the title is stored
    in the documentation."""
    id = 'C2017'
    apply_to = COMM_EXT
    apply_ctxt = CheckContext(branch=['default'])
    kwd = ['TITRE', ]
    search = search_msg("^ *# *(?P<main>(%s) .*)" % '|'.join(kwd))


class LicenseNotFound(FileContentCat, COMM.LicenseNotFound):

    """Summary with GPL is required
    The copyright and GPL summary are required. Even if there is an external
    copyright, it must mention www.code-aster.org"""
    apply_to = COMM_EXT
    id = "C2006"


class EDFCopyright(FileContentCat, COMM.EDFCopyright):

    """EDF R&D Copyright not found
    Even if there is a copyright to another company than EDF R&D, the source
    has been changed for Code_Aster conformance and/or version management.
    Example:

       - Copyright 2008 Company ABC
       - Copyright 2008-2013 EDF R&D www.code-aster.org
    """
    apply_to = COMM_EXT
    id = "C2007"


class InvalidCharacter(FileContentCat, COMM.InvalidCharacter):

    r"""Invalid character, unexpected tabulation '\t'
    Tabulation character is not allowed."""
    apply_to = COMM_EXT
    id = "C2008"


class RequiredKeyword(FileContentCat, GenericMsg):

    """Required keyword in a command file
    A testcase must use one of operators DEBUT or POURSUITE and the keyword
    CODE."""
    apply_to = COMM_EXT
    apply_ctxt = CheckContext(reponame=['src'], branch='^(default|asterxx)')
    id = 'C2011'
    _debut = re.compile('^DEBUT', re.M)
    _debcod = re.compile("^DEBUT *\([^\#]*CODE *= *_F\(", re.M)
    _deberr = """\"CODE=_F(...)\""""
    _pours = re.compile('^POURSUITE', re.M)
    _poucod = re.compile("^POURSUITE *\([^\#]*CODE *= *[\'\"]OUI", re.M)
    _pouerr = """\"CODE='OUI'\""""

    def search(self, txt):
        """Check a command file for required keywords"""
        err = []
        if self._debut.search(txt):
            if not self._debcod.search(txt):
                err.append(": %s not found in DEBUT" % self._deberr)
        elif self._pours.search(txt):
            if not self._poucod.search(txt):
                err.append(": %s not found in POURSUITE" % self._pouerr)
        elif "asterxx" in self.ctxt.branch:
            pass
        else:
            err.append(": neither DEBUT or POURSUITE found")
        return err

CHECK_LIST = checkers_from_context(globals(), TextMsg)
