# coding=utf-8

"""commands of the Code_Aster extension"""

import os
import os.path as osp
import shutil
import tempfile
from functools import partial
from subprocess import PIPE, STDOUT, Popen, call

from mercurial import cmdutil, util

from api_doc import changelog_doc
from api_roundup import check_status, get_connection, get_documents
from aslint.api_asrun import check_asrun_bin
from aslint.base_checkers import Context, Report, checkcontext
from aslint.check_files import (build_testcase_list, check_diff, check_files,
                                exclude_test, filter_list, read_waf_parameters,
                                run_testcase, run_waf_command)
from aslint.check_inst import check_install
from aslint.config import ASCFG, SMCFG, TOOLCFG, XXCFG
from aslint.decorators import log_as_important, stop_on_failure
from aslint.filetype import filter_dict, filter_names
from aslint.i18n import _
from aslint.logger import ErrorCode, logger
from aslint.string_utils import convert, to_unicode
from aslint.utils import convert_to_addr

from .ext_utils import (check_for_updates, get_changed_files, get_description,
                        get_diff, get_hgpath, get_issues_from_descr,
                        get_reference_ancestor, get_repo_by_rev, get_repo_name,
                        get_solved_issues, has_uncommitted_changes,
                        parent_is_last, sendmail, shortrev, shortrev_s,
                        valid_branch)
from .request_queue import QueueError, Request, RequestQueue, requestkey
from .hooks.generic import OK

cmdtable = {}
command = cmdutil.command(cmdtable)

# we communicate only using the logger object in commands (not hooks)
#   ui.debug --> logger.debug
#   ui.write, ui.status, ui.note --> logger.info
#   ui.warn --> logger.warn
globopts = [
    ('n', 'dry-run', False,
     _("does nothing, just show what it would be done")),
    ('', 'local', False,
     _("do not make remote connections, updates will not be downloaded, "
       "no request for integration will be submitted")),
    ('r', 'rev', '',
     _("last ancestor revision already integrated. "
       "For submit, only use if --local "
       "is enabled because it can not be retreived from the remote "
       "repository"), _('REV')),
]
testopts = [
    ('', 'resutest', '',
     _("directory to write the results of the testcases"), _('DIR')),
]
msgopts = [
    ('m', 'message', '',
     _("message sent to the developer"), _('TEXT')),
]


@command("submit",
         globopts + testopts,
         _('[--local -r REV] [OPTION]...'))
def submit(ui, repo, *args, **opts):
    """request the integration of the parent revision

    Request a verification of Code_Aster source files at the parent revision
    and, if there is no error, submit a request for integration in the main
    repository.

    The successive tasks are:

    - check for new updates for the current repository and the 'devtools' one

    - check that there is not already a pending request in the same branch

    - build the executable and catalog using waf

    - check the installation made by waf

    - check the source files and make a report (*)

    - run a minimal list of testcases + those that have been added or changed

    - if all the testcases are OK, the revision is pushed to the server and a
      request for integration is sent

    - if one testcase fails, the request for integration is cancelled

    - notify the result by mail

    (*) if warnings are emitted at this step, the developer is prompted to
        interrupt the process to fix them or to continue but the integration
        could be rejected.

    When --local option is used, it only checks the source files. It does not
    check/submit a request for integration, neither for new updates.

    The process is aborted when an error is raised.

    Returns 0 on success, != 0 if an error occurred.
    """
    reponame = get_repo_by_rev(repo)
    if reponame == "devtools":
        class_ = SubmitDevtools
    elif reponame in ("data", "src", "validation"):
        class_ = SubmitAsterLegacy
    elif reponame in ("data-xx", "src-xx", "validation-xx"):
        class_ = SubmitAsterXX
    elif reponame == "salome-codeaster-study":
        class_ = SubmitAsterStudy
    elif reponame == "salome-meca-rtool":
        class_ = SubmitRTool
    else:
        raise util.Abort(_("'submit' is not supported in this repository"))
    return class_(ui, repo, *args, **opts).run()

requestOptions = [
    ('', 'start', False,
     _("Initiate the integration of the revision REV and notify the "
       "developer that an integrator is working on its proposal. The status "
       "of the request is changed to RUN."), ''),
    ('', 'startall', False,
     _("Apply start on all pending requests."), ''),
    ('', 'check', False,
     _("Run the different checkings similar to the ones run at the submission."), ''),
    ('', 'test', False,
     _("Mark the parent revision to be tested. If the changesets have been "
       "rearranged, the parent revision is different of REV, in this case, "
       "it is mentionned in the commit message. The status of the request "
       "is changed to TEST (needs pass testcases). "
       "The developer is notified."), ''),
    ('', 'testall', False,
     _("Apply test on all requests in the RUN status."), ''),
    ('', 'close', False,
     _("The request for integration is closed because testcases passed "
       "with success. Called automatically when testcases are finished. "
       "The developer is notified of the closure."), ''),
    ('', 'closeall', False,
     _("Apply close on all requests in the TEST status."), ''),
    ('', 'refuse', False,
     _("The request for integration is closed because testcases failed. "
       "Called automatically when testcases are finished. "
       "The developer is notified of the closure."), ''),
    ('', 'differ', False,
     _("Differ the integration of the revision REV and notify the developer "
       "that its proposal will be analyzed later. The status of the request "
       "is reset to PEND."), ''),
    ('', 'cancel', False,
     _("The request for integration is cancelled, the revision won't be "
       "integrated until a new request will be submitted. "
       "The developer is notified of the removal."), ''),
    ('', 'status', False,
     _("Return the current status of the request for integration of a "
       "revision or of all registered requests."), ''),
    # ('', 'test', '',
    #  _(""), ''),
]

requestActions = [opt[1] for opt in requestOptions]


@command("request",
         requestOptions + globopts + msgopts,
         _('[options] [REV...]'))
def request(ui, repo, *args, **opts):
    """set of commands to manage the workflow of a request for integration

    This command is dedicated to the administrators that have to integrate a
    candidate revision. It offers several elementary actions (see options).

    Note:
        REV is compulsory for --start, --differ, --test, --close, --refuse
        and --cancel, it is the revision proposed for integration.
        --close and --refuse should only be called by script after the testcases
        run.

    The workflow is (actions are between [])::

                    [check]                                                     .
            [start]   /\                                                       .
           ------->   \/   [test]         [close]                              .
     PEND  <-------  RUN  ------->  TEST  ------> closed, removed from the queue
        \   [differ]     \               \                                     .
         \                \ [cancel]      \ [refuse]                           .
          \ [cancel]       \               \                                   .
           `-> removed      `-> removed     `-> removed                        .
    """
    return Integration(ui, repo, *args, **opts).run()


MAX_MESSAGE_SIZE = int(1e6)

HEADER_DIFF = _(u"""
The developer (username=%(dvp)s, address=%(address)s) submitted a
request for integration of the revision %(wrkrev)s in the Code_Aster repository.

You are registered as the person in charge of the following files.

If you don't agree with these changes, you must warn the integrators and the
developer not to integrate them as soon as possible.

URL of the submitted changeset: %(url)s/rev/%(wrkrev)s

Note that its ancestors are also part of the changes.

If you want to rebuild this entire diff file in your local repository, you have
to pull this revision and type:

    hg pull --rev %(wrkrev)s %(url)s
    hg diff --rev %(refrev)s --rev %(wrkrev)s
""")

SUBMIT_TEMPLATE = _(u"""
Important messages of the submission:

%(msg)s

Full log content was written to: %(logf)s
""")

SEPAR = """
-----
"""


class CandidateRevisionCheck(object):

    """Offer several unit command to check a revision, to submit it,
    to integrate it

    Attributes:
        ui, repo: ui and repository objects passed from mercurial
        reponame: name of the repository
        rev: revision to work on
        branch: branch name
        _queue: RequestQueue object
        refrev: reference revision
        dvp: username of the developer
        _report: Report object of the checkings
        _changes: dict of changes by type between rev and refrev
        c4che: waf cache directory where to read parameters
    """
    _interrupt_message = _("command interrupted because of errors")
    _continue_message = _("do you want to continue (y/n) ?")
    _cancel_message = _("command cancelled")

    def __init__(self, ui, repo, *args, **opts):
        """Initialization"""
        self._name = 'not_set'
        self.ui = ui
        logger.setLevel(_loglevel(ui))
        self._notify_on_error = False
        self.repo = repo
        self._dry = opts.get('dry_run')
        self._local = opts.get('local')
        self._refancestor = opts.get('rev')
        self.wrkrev = None
        self.reponame = None
        self.wrkbranch = None
        self.refbranch = None
        self._srcdir = None
        self._queue = None
        self.refrev = None
        self.dvp = self.get_cfg('username')
        self._report = None
        self._changes = None
        self.c4che = None

    @classmethod
    def get_cfg(self, key, default=None):
        """Get a value of a parameter from the configuration."""
        return ASCFG.get(key, default)

    def _dump_log(self):
        """Dump the log content into a file"""
        return logger.write_log(self._name, path=osp.dirname(self.repo.root))

    def _error(self, message):
        """Log the message error and raise the hg Abort exception"""
        logger.error(message)
        self._dump_log()
        raise util.Abort(self._interrupt_message)

    def _check_error(self):
        """Interrupt the process if an error occurred"""
        if not logger.is_error():
            return
        if self._notify_on_error and not self._local:
            sendmail('user', self._interrupt_message)
        raise util.Abort(self._interrupt_message)

    def confirm(self):
        """Ask confirmation"""
        if self._dry:
            logger.warn(_("skip confirmation"))
        else:
            answ = self.ui.prompt(self._continue_message, default='n')
            if answ.lower() not in ('y', 'o'):
                self._error(self._cancel_message)

    def set_working_revision(self):
        """Set the working revision and branch"""
        self.wrkrev = self.repo['.']
        self.wrkbranch = self.wrkrev.branch()
        self.refbranch = self.refbranch or self.wrkbranch
        checkcontext.branch = self.refbranch

    def set_parameters(self, need_changes=True):
        """Set the parameters of the repository"""
        try:
            self.reponame = get_repo_name(self.repo)
            checkcontext.reponame = self.reponame
        except KeyError:
            if not self.get_cfg('admin._test'):
                self._error(_("not in a Code_Aster repository: {0}")
                            .format(self.repo.root))
        if self._refancestor:
            if self._name == 'submit' and not self._local:
                self._error(_("'--rev' must only be used with '--local'"))
            self.refrev = self._refancestor
        else:
            if not self._local:
                self.refrev = self.get_reference_revision()
        if not self._local:
            if not self.valid_branch():
                self._error(_("integration is not allowed in the branch '{0}'")
                            .format(self.wrkbranch))
        if not self.refrev:
            self._error(_("without remote connection (--local enabled), "
                          "you must provide the last ancestor revision "
                          "already integrated using --rev"))
        # 'wrkrev' must be a descendant of 'refrev'
        ref = self.repo[self.refrev]
        if self.wrkrev.ancestor(ref) != ref:
            self._error(_("The submitted revision must be a descendant of the "
                          "reference head."))
        if need_changes and self.wrkrev.hex() == self.refrev:
            self._error(_("You are submitting the reference revision!"))
        if self.reponame == 'src':
            srcdir = self.repo.root
        else:
            srcdir = osp.join(osp.dirname(self.repo.root), 'src')
        if not osp.join(srcdir):
            self._error(_("source repository not found: '{0}'").format(srcdir))
        self._srcdir = srcdir

    def init_report(self):
        """Initialize the Report object"""
        context = Context()
        context.username = self.dvp
        context.repository = self.reponame
        context.submission = True
        self._report = Report(context)

    def init_queue(self, notify=None):
        """Create the RequestQueue object"""
        if self._queue is not None:
            return
        if notify is None:
            notify = not self._dry
        try:
            self._queue = RequestQueue(notify=notify)
            self._queue.init()
        except QueueError, exc:
            logger.error(exc.message)
        self._check_error()

    def check_children(self):
        """Check for children of the working revision"""
        if self.wrkrev.children():
            logger.warn(_("the submitted revision has children"))

    def check_uncommitted(self):
        """Check for uncommitted changes"""
        if has_uncommitted_changes(self.repo):
            if not self.get_cfg('admin._test'):
                self._error(_("there are uncommitted changes. Please commit "
                              "or revert first."))
            else:
                logger.warn(_("uncommitted changes, skipped in TEST mode"))
        elif has_uncommitted_changes(self.repo, unknown=True):
            logger.warn(_("there are untracked files"))
            self.confirm()
            logger.cancel_warning()

    def check_repositories_updates(self):
        """Check for updates of the required repositories"""
        if not self._local:
            opts = dict(source=self.get_cfg('admin_pull_uri'),
                        branch=[self.refbranch])
            _check_for_updates(self.reponame, self.repo.root, **opts)
            self._check_error()
        else:
            logger.warn(_("skipping check for updates"))

    @stop_on_failure
    @log_as_important
    def build_changes_dict(self):
        """changes with reference revision"""
        wrkrev = self.wrkrev
        refrev = self.refrev
        revs = [refrev, wrkrev.hex()]
        strref = shortrev_s(refrev)
        strrev = str(wrkrev.rev()) + ':' + shortrev(wrkrev)
        if wrkrev != self.repo['.']:
            self._error(_("run 'hg update {0}' to update the working directory "
                          "to the revision to integrate").format(wrkrev.rev()))
        logger.title(_("getting changes between revisions '{0}' and '{1}'")
                     .format(strref, strrev))
        # 'wrkrev' must be a descendant of 'refrev'
        if wrkrev.ancestor(self.repo[refrev]) != self.repo[refrev]:
            self._error(_("'{0}' must be a descendant of '{1}'")
                        .format(strrev, strref))
        dchg = get_changed_files(self.ui, self.repo, revs)
        if len(dchg['!']):
            # should not happen after has_uncommitted_changes() checkings
            self._error(_("there are missing files (deleted by non-hg "
                          "command, but still tracked, "
                          "check with 'hg status')"))
        if dchg['msg']:
            logger.info(dchg['msg'])
        self._changes = filter_dict(dchg)
        return logger.errcode

    @stop_on_failure
    @log_as_important
    def check_message_descr(self):
        """Check the commits description"""
        wrkrev = self.wrkrev
        refrev = self.refrev
        strref = shortrev_s(refrev)
        strrev = str(wrkrev.rev()) + ':' + shortrev(wrkrev)
        logger.title(_("checking descriptions between revisions '{0}' and '{1}'")
                     .format(strref, strrev))
        descr = get_description(self.ui, self.repo, strref, wrkrev.hex())
        if len(descr) == 0:
            logger.error(_("there is no revision to submit"))
        for msg in descr:
            _check_description(msg)
        return logger.errcode

    @log_as_important
    def report_on_new(self):
        """limit the report to the new files"""
        logger.add_cr()
        logger.title(_("report limited to the added files"))
        context = Context()
        context.username = self.dvp
        context.repository = self.reponame
        report = Report(context)
        new = []
        for flist in self._changes.values():
            new.extend(filter_names(flist, 'A'))
        for fname, msglist in self._report.with_msg_iter():
            if fname in new:
                report.set(fname, msglist)
        logger.info(report.to_text())
        report.report_summary()
        return report

    def accept_special_issue(self, issue_id):
        """Allow to accept issue even if the status is not as expected."""
        return False

    @stop_on_failure
    @log_as_important
    def check_issues_status(self, expected='valide_EDA', logfunc=logger.error):
        """checking the status of the solved issues"""
        wrkrev = self.wrkrev
        refrev = self.refrev
        revs = [refrev, wrkrev.hex()]
        strref = shortrev_s(refrev)
        strrev = str(wrkrev.rev()) + ':' + shortrev(wrkrev)
        logger.title(_("getting solved issues between '{0}' and '{1}'")
                     .format(strref, strrev))
        if self._local:
            logger.warn(
                _("status of issues can not be checked in --local mode"))
            return logger.errcode
        lrex, lbb = get_solved_issues(self.ui, self.repo, *revs)[1:]
        rex = get_connection()
        if rex is None:
            logger.warn(_("please check that their status is '{0}'")
                        .format(expected))
            for num in lrex:
                logger.warn(_("ensure that the issue #{0:d} is validated "
                              "and that related documents have been changed")
                            .format(num))
            return logger.errcode
        for num in lrex:
            issue_id = "issue" + str(num)
            issue = rex.display(issue_id)
            check_status(issue_id, issue, expected, rex,
                         verbose=True,
                         force=self.accept_special_issue(num),
                         logfunc=logfunc)
        for num in lbb:
            logger.warn(
                _("ensure that the issue #%d is solved on bitbucket"), num)
        if not logger.is_error():
            self.check_expected_documents(lrex)
        return logger.errcode

    @stop_on_failure
    @log_as_important
    def check_expected_documents(self, issues):
        """Check that expected documents have been committed.

        Arguments:
            list[int]: List of issues numbers.
        """
        logger.title(_("check that expected documents have been committed"))
        rex = get_connection()
        for num in issues:
            docs = get_documents(str(num), rex)
            if not docs:
                logger.info(_("no changes needed for #{0}").format(num))
            for key in docs:
                try:
                    descr = " ".join(changelog_doc(key))
                except ValueError as exc:
                    descr = ""
                done = get_issues_from_descr(descr, bracket=False)
                msg = _("issue #{0} needs changes in {1}: {2}")
                if num in done:
                    logger.info(msg.format(num, key, _("found")))
                else:
                    logger.warn(msg.format(num, key,
                                           _("issue number not found in "
                                             "document history")))
        return logger.errcode

    @stop_on_failure
    @log_as_important
    def push_revision(self, dest):
        """Push the working revision"""
        if not self._local:
            act = 'push'
            if self._dry:
                act = 'outgoing'
            logger.title(_("pushing the revision %s to the remote server"),
                            self.wrkrev.hex())
            srev = str(self.wrkrev.rev())
            iret = call([self.get_cfg('hg_bin'), act, '--force',
                        '--rev', srev, dest])
            if iret == 0:
                logger.info(_("push ended successfully"))
            else:
                logger.warn(_("hg push exits with code %s"), iret)
                logger.warn(_("no revision pushed, maybe already there"))
                logger.warn(_("Please force the push by using:") + '\n' +
                            "hg push --force --rev %s %s", srev, dest)
        else:
            logger.warn(_("the revision is not push to the remote server"))
        return logger.errcode

    def check_init(self):
        """Execute the elementary tasks"""
        try:
            ASCFG.check_init()
        except AssertionError, exc:
            self._error(_("incorrect configuration:\n%s") % str(exc))
        if self.get_cfg('admin._test'):
            logger.warn(_("running is TEST mode"))
            self.confirm()

    def get_reference_revision(self):
        """Return the reference revision."""
        refrev = get_reference_ancestor(self.ui, self.repo,
                                        self.wrkrev, self.refbranch,
                                        self.get_cfg('admin_pull_uri'))
        if not refrev:
            self._error(_("can not find the last integrated revision"))
        return refrev

    def check_single_head(self):
        """Check for single heads in branch."""
        from hooks.generic import single_head_per_branch_hook
        checker = single_head_per_branch_hook()
        return checker(self.ui, self.repo, self.wrkrev) == OK

    def check_branch(self, repo, branch): # pragma pylint: disable=unused-argument
        """Check branch validity."""
        raise NotImplementedError("must be subclassed")

    def valid_branch(self):
        """Check branch validity."""
        return (self.check_branch(self.repo, self.wrkbranch) and
                self.check_single_head())


class CandidateRevisionCheckAster(CandidateRevisionCheck):

    """Abstract class for code_aster repositories.
    """

    def check_repositories_updates(self):
        """Check for updates of the required repositories"""
        if not self._local:
            dev = "devtools"
            branch = "default"
            _check_for_updates(dev, update=True, branch=[branch])
            if not parent_is_last(self.get_cfg('devtools_root')):
                logger.error((_("you can execute:") +
                             " cd {0} && hg up -C {2} && cd {1}")
                             .format(self.get_cfg('devtools_root'), os.getcwdu(),
                                     branch))
                self._error(
                    _("'{0}' must be updated to the last revision "
                      "of branch '{1}'").format(dev, branch))
            opts = dict(source=self.get_cfg('admin_pull_uri'),
                        branch=[self.refbranch])
            _check_for_updates(self.reponame, self.repo.root, **opts)
            if self.reponame != 'src':
                _check_for_updates('src', self._srcdir, **opts)
            self._check_error()
        else:
            logger.warn(_("skipping check for updates"))

    @stop_on_failure
    def check_source_files(self, askconfirm=True):
        """checking files"""
        prev = os.getcwd()
        os.chdir(self.repo.root)
        check_files(self._report, self._changes, self.c4che,
                    srcdir=self.repo.root, dry_run=self._dry)
        logger.add_cr()
        self.check_diff_files()
        logger.add_cr()
        logger.info(self._report.to_text())
        self._report.report_summary()
        os.chdir(prev)
        if not askconfirm:
            logger.cancel_error()
            return logger.errcode
        if logger.is_warn():
            logger.warn(_("warnings have been emitted during check, try to fix "
                          "most of them before continuing"))
            self.confirm()
        elif logger.is_error():
            logger.error(_("Errors have been emitted during check, "
                           "you must fix them."))
            if self.get_cfg('check.submit', 'error') == 'warn':
                self.confirm()
                logger.cancel_error()
        return logger.errcode

    def check_diff_files(self):
        """checking the diff files"""
        prev = os.getcwd()
        os.chdir(self.repo.root)
        wrkrev = self.wrkrev
        refrev = self.refrev
        revs = [refrev, wrkrev.hex()]
        diffs = {}
        for typ, chg in self._changes.items():
            lfiles = filter_names(chg, status='M')
            diffs[typ] = {}
            for fname in lfiles:
                adiff = get_diff(self.ui, self.repo, revs, [fname])
                diffs[typ][fname] = adiff
        check_diff(self._report, diffs)
        os.chdir(prev)

    @stop_on_failure
    def check_waf_build(self):
        """Check waf build"""
        logger.title(_("building the binaries using waf"))
        have_mpi = self.c4che.get('HAVE_MPI')
        if have_mpi:
            self._error(_("a sequential version is necessary; "
                          "use 'waf.builddir' to select another build."))
        prefix = self.c4che['PREFIX']
        ipref = osp.abspath(prefix).split(os.sep, 2)[1]
        isrc = osp.abspath(self._srcdir).split(os.sep, 2)[1]
        opts = []
        if ipref != isrc and not self.c4che['install_tests']:
            logger.info(_("source and installation directories seem to be on "
                          "different filesystems."))
            opts.append('--install-tests')
        return run_waf_command(self._srcdir,
                               osp.join(self._srcdir, self.get_cfg('waf.bin')),
                               'install',
                               options=opts,
                               env=self.get_cfg('waf.env'),
                               dry_run=self._dry)

    def init_waf_param(self):
        """Read and store the waf parameters"""
        wafc4che = osp.join(self._srcdir, self.get_cfg('waf.builddir'),
                            'c4che', 'release_cache.py')
        logger.debug('waf c4che: %s', wafc4che)
        self.c4che = read_waf_parameters(wafc4che)
        if not self.c4che:
            self._error(_("no such file: '{0}', please run '{1} configure'. "
                          "If you are using a specific version you can "
                          "adjust the 'waf.builddir' parameter in the [aster] "
                          "section of your ~/.hgrc file.")
                        .format(wafc4che, self.get_cfg('waf.bin')))

    @stop_on_failure
    def check_waf_install(self):
        """Check waf installation"""
        return check_install(self.c4che, dry_run=self._dry)


class AbstractSubmit(CandidateRevisionCheck):

    """Check code_aster source files at the parent revision
    and, in case of success, submit a request for integration.

    Attributes:
        command_automerge (list): Command (as list) for automerge checking.
        command_lint (list): Command (as list) for source files checking.
        command_tests (list[list]): List of commands (each one as list) for
            running unittests.
    """
    _interrupt_message = _("submission interrupted because of errors")
    _cancel_message = _("submission cancelled")
    command_automerge = None
    command_lint = None
    command_tests = None

    def __init__(self, ui, repo, *args, **opts):
        """Initialization"""
        super(AbstractSubmit, self).__init__(ui, repo, *args, **opts)
        self._name = 'submit'
        self._notify_on_error = True
        if len(args) > 0:
            self._error(_("no argument expected"))

    @stop_on_failure
    def check_existing_request(self):
        """check if there is already a submitted request"""
        if self._local:
            logger.warn(_("no request for integration will be submitted, only "
                          "check the source files"))
            return logger.errcode
        if not self.dvp:
            logger.error(_("you must fill the 'username' parameter in the "
                           "[ui] section of your ~/.hgrc file"))
        self.init_queue(notify=False)
        self._check_error()
        reqkey = requestkey(self.reponame, self.wrkbranch, self.dvp)
        req = self._queue.get(reqkey)
        if req is not None:
            logger.warn(_("you have already submitted a request for "
                          "integration of the revision %s"), req.shortnode)
            if req.status != 'PEND':
                logger.error(_("the integration is started, you must contact "
                               "the administrators to cancel it"))
            else:
                logger.warn(_("if you continue, the previously submitted "
                              "revisions won't be integrated."))
                self.confirm()
                logger.cancel_warning()
                if not self._dry:
                    logger.update_from_status(self._queue.cancel(reqkey))
                else:
                    logger.info(_("(DRY RUN) cancelling the submission"))
        req = self._queue.get_by_node(self.wrkrev.hex())
        if req is not None:
            logger.error(_("this revision has already been submitted by "
                           "another developer (%s)"), req.username)
        return logger.errcode

    @stop_on_failure
    def submit_revision(self):
        """Submit the integration request"""
        if self._queue is not None:
            addr = convert_to_addr(self.get_cfg('notify.user'))
            addr = addr and addr[0] or ''
            req = Request(self.reponame, self.wrkbranch, self.dvp,
                          None, self.wrkrev.hex(), email=addr)
            if not self._dry:
                iret = self._queue.submit(req)
                if iret != 0:
                    logger.error(_("PLEASE FORWARD THIS MESSAGE TO "
                                   "AN ADMINISTRATOR"))
            else:
                logger.info(_("(DRY RUN) submit request for %s"),
                            req.username)
        else:
            logger.warn(_("your request for integration has not been "
                          "submitted"))
        return logger.errcode

    @stop_on_failure
    def notify_responsible(self):
        """Notify the responsible of the changed files"""
        if self._local:
            logger.warn(_("do not send notification to persons in charge"))
            return logger.errcode
        from aslint.common_checkers import PersonInCharge
        assert self._report is not None
        res = self._report.report_by_type()
        nbchg = res[1].get(PersonInCharge.id, 0)
        if nbchg < 1:
            logger.info(_("no person in charge of the changed modules"))
            return logger.errcode
        logger.info(_(u"%d source files have a person in charge"), nbchg)
        # list of files by person in charge
        fname_pic = {}
        for fname, msglist in self._report.with_msg_iter(not_empty=True):
            pic = [msg.result['main'] for msg in msglist
                   if isinstance(msg, PersonInCharge)]
            list_addr = convert_to_addr(pic)
            if not list_addr:
                continue
            for add in list_addr:
                fname_pic[add] = fname_pic.get(add, []) + [fname, ]
        # send diff to dvp
        wrkrev = self.wrkrev
        refrev = self.refrev
        revs = [refrev, wrkrev.hex()]
        if self.get_cfg('admin_push_uri'):
            url = get_hgpath(self.ui, self.repo, self.get_cfg('admin_push_uri'))
        else:
            url = None
        dest = convert_to_addr(self.get_cfg('notify.user'))
        dest = dest and dest[0] or ''
        header = HEADER_DIFF % {
            'dvp': self.dvp,
            'address': dest,
            'wrkrev': shortrev(wrkrev),
            'refrev': shortrev_s(refrev),
            'url': url}
        trunc = _(u"The entire diff is too long. It has been truncated.")
        for add, lfiles in fname_pic.items():
            if '@' not in add:
                logger.warn(_(u"invalid address '{0}' in files: {1}")
                            .format(add, lfiles))
                continue
            if add.strip() == dest.strip():
                logger.info(_("no mail sent for the files which you are the "
                              "person in charge: {0}").format(lfiles))
                continue
            text = [header]
            diffs = get_diff(self.ui, self.repo, revs, lfiles)
            if len(diffs) > MAX_MESSAGE_SIZE:
                text.append(trunc)
                diffs = diffs[:MAX_MESSAGE_SIZE]
                diffs += os.linesep + "[...]"
            text.extend([SEPAR, _(u"List of changed files:"), ""])
            text.extend(lfiles)
            text.append(SEPAR)
            text.append(diffs)
            text = [convert(part) for part in text]
            text = os.linesep.join(text)
            sendmail(add,
                     "[submit@{host}] %s is changing files you are responsible"
                     % self.dvp,
                     text, dry_run=self._dry)
        return logger.errcode

    def _check_command(self, title, command):
        """Execute a shell command"""
        if not command:
            return logger.errcode
        errcode = ErrorCode()
        logger.title(title)
        scmd = command
        if isinstance(command, (list, tuple)):
            scmd = " ".join(command)
        logger.debug(_("Executing: {0}").format(scmd))
        proc = Popen(command, stdout=PIPE, stderr=STDOUT, bufsize=1)
        for line in iter(proc.stdout.readline, ''):
            logger.info(line.rstrip())
        proc.stdout.close()
        returncode = proc.wait()
        errcode.from_status(returncode)
        if errcode.is_ok():
            logger.info(_("'{0}' ended successfully").format(title))
        else:
            self._error(_("'{0}' failed.\nCommand line: {1}")
                        .format(title, command))
        return logger.errcode

    @stop_on_failure
    @log_as_important
    def check_automerge(self):
        """Check merge without conflict."""
        return self._check_command(_("checking for automatic merge"),
                                   self.command_automerge)

    @stop_on_failure
    def check_source_files(self, askconfirm=True):
        """checking files"""
        return self._check_command(_("checking for source files"),
                                   self.command_lint)

    @stop_on_failure
    @log_as_important
    def check_testcase_run(self):
        """running testcases"""
        cmds = self.command_tests
        if not cmds:
            return logger.errcode
        elif not isinstance(cmds[0], list):
            cmds = [cmds]
        for cmd in cmds:
            self._check_command(_("running unittests"), cmd)
        return logger.errcode

    def end_message(self):
        """The end"""
        if self._local:
            logger.warn(_("only check files has been done"))
        elif not self._dry:
            if not logger.is_ok():
                logger.warn(_("the integration may be rejected because of "
                              "previous alarms"))
            else:
                logger.info(_("the integration should be accepted"))
        logf = self._dump_log()
        if not self._local:
            subj = _("[submit@{{host}}] request for integration submitted "
                     "({0}/{1}) {2}").format(self.reponame, self.wrkbranch,
                                             shortrev(self.wrkrev))
            sendmail(['user', 'admins'], subj,
                     SUBMIT_TEMPLATE % {
                     'msg': to_unicode(logger.get_important_text()),
                     'logf': logf},
                     dry_run=self._dry)

    def run(self):
        """Execute the elementary tasks"""
        super(AbstractSubmit, self).check_init()
        self.set_working_revision()
        self.set_parameters()
        logger.title(_("asking integration in branch '{0}' of the repository"
                       " '{1}'").format(self.wrkbranch, self.reponame))
        logger.info(_("reference revision is %s"), self.refrev[:12])
        self.check_uncommitted()
        self.check_children()
        self.check_repositories_updates()
        self.check_automerge()
        self.check_message_descr()
        self.check_issues_status()
        self.init_report()
        self.check_source_files()
        self.check_testcase_run()
        self.push_revision(dest=self.get_cfg('admin_ci_uri'))
        self.notify_responsible()
        self.end_message()
        return 0


class AbstractSubmitAster(CandidateRevisionCheckAster, AbstractSubmit):

    """Check code_aster source files at the parent revision
    and, in case of success, submit changesets for continuous integration.
    """

    def __init__(self, ui, repo, *args, **opts):
        """Initialization"""
        super(AbstractSubmitAster, self).__init__(ui, repo, *args, **opts)
        self._resutest = opts.get("resutest")

    @stop_on_failure
    def check_testcase_run(self):
        """running testcases"""
        try:
            versdir = self.c4che['ASTERDATADIR']
        except (TypeError, KeyError):
            logger.error(_("waf parameters are required"))
            return logger.errcode
        params = {
            'version': versdir,
            'have_mpi': self.c4che.get('HAVE_MPI', 0) == 1,
            'nomjob': 'test_%s' % shortrev(self.wrkrev),
        }
        if self._resutest:
            params['resu_test'] = self._resutest
        if self.reponame != 'data':
            testdir = osp.join(self.repo.root, 'astest')
        else:
            testdir = osp.join(self._srcdir, 'astest')
        flist = build_testcase_list(self._changes, testdir, repo=self.reponame)
        exclude_test(flist, 'list.exclude.submit')
        exclude_test(flist, 'list.exclude')
        exclude_test(flist, 'list.not_rerun')
        filter_list(flist, params, params['have_mpi'],
                    filters=['--filter="validation" not in testlist'])

        params['list'] = flist
        errcode = run_testcase(params, dry_run=self._dry)[0]
        if errcode.is_error():
            logger.warn(_("Your request should be refused, but if you think "
                          "that it will be accepted anyway, you can continue."))
            self.confirm()
            logger.cancel_error()
        return logger.errcode

    def run(self):
        """Execute the elementary tasks"""
        super(AbstractSubmitAster, self).check_init()
        self.set_working_revision()
        self.set_parameters()
        logger.title(_("asking integration in branch '{0}' of the repository"
                       " '{1}'").format(self.wrkbranch, self.reponame))
        logger.info(_("reference revision is %s"), self.refrev[:12])
        if self.reponame != 'src':
            logger.warn(_("please check that the current working directory "
                          "of 'src' is consistent with this repository "
                          "because it will be used to run testcases"))
            self.confirm()
            logger.cancel_warning()
        check_asrun_bin()
        self.check_uncommitted()
        self.check_children()
        self.check_repositories_updates()
        self.check_existing_request()
        self.build_changes_dict()
        self.check_message_descr()
        self.check_issues_status()
        self.init_waf_param()
        self.check_waf_build()
        self.check_waf_install()
        self.init_report()
        self.check_source_files()
        self.check_testcase_run()
        self.push_revision(dest=self.get_cfg('admin_ci_uri'))
        self.submit_revision()
        self.notify_responsible()
        self.end_message()
        return 0


class SubmitAsterLegacy(AbstractSubmitAster):

    """Check code_aster source files at the parent revision
    and, in case of success, submit changesets for continuous integration.
    """

    @staticmethod
    def check_single_head():
        """Check single heads in branch."""
        # multiple heads allowed in code_aster legacy repositories
        return True

    def check_branch(self, repo, branch):
        """Check branch validity."""
        return valid_branch(repo, branch)


class SubmitAsterXX(AbstractSubmitAster):

    """Check code_aster++ source files at the parent revision
    and, in case of success, submit changesets for continuous integration.
    """
    command_tests = None # defined by constructor

    def __init__(self, ui, repo, *args, **opts):
        """Initialization"""
        super(SubmitAsterXX, self).__init__(ui, repo, *args, **opts)
        self.refbranch = "asterxx"

        resudir = tempfile.mkdtemp(prefix='resutest_')
        self.command_tests = [
            [self.get_cfg('waf.bin'), "doc"],
            ["./check_docs.sh", "--waf", self.get_cfg('waf.bin'),
             "--builddir", self.get_cfg("waf.builddir")],
            ["run_testcases", "--root=..", "--testlist=asterxx",
             "--builddir", self.get_cfg("waf.builddir"),
             "--resutest", resudir, "--clean"],
        ]

    @classmethod
    def get_cfg(self, key, default=None):
        """Get a value of a parameter from the configuration."""
        # Force sequential version for submit
        XXCFG.waf_bin = "./waf_std"
        XXCFG.waf_builddir = "build/std"
        return XXCFG.get(key, default)

    def check_branch(self, dummy, branch):
        """Check branch validity."""
        from hooks.generic import branch_checker
        return branch_checker('^asterxx/')(branch) == OK

    def accept_special_issue(self, issue_id):
        """Allow to accept issue even if the status is not as expected."""
        return issue_id in (23037, )

    def submit_revision(self):
        """Does nothing since integration is automatic in this repository."""
        pass

    @stop_on_failure
    @log_as_important
    def check_testcase_run(self):
        """running testcases"""
        if self.reponame != "src":
            return logger.errcode

        return AbstractSubmit.check_testcase_run(self)

    @stop_on_failure
    def notify_responsible(self):
        """Notify the responsible of the changed files,
        *temporarly disabled*."""
        logger.warn(_("do not send notification to persons in charge"))
        return logger.errcode


class SubmitDevtools(AbstractSubmit):

    """Check devtools source files at the parent revision
    and, in case of success, submit changesets for continuous integration.
    """
    command_automerge = ["./share/test/check_automerge.sh"]
    command_tests = ["./share/test/check_tests.sh", "--coverage"]

    def __init__(self, ui, repo, *args, **opts):
        """Initialization"""
        super(SubmitDevtools, self).__init__(ui, repo, *args, **opts)
        self.refbranch = "default"

    @classmethod
    def get_cfg(self, key, default=None):
        """Get a value of a parameter from the configuration."""
        return TOOLCFG.get(key, default)

    def check_issues_status(self, expected='valide_EDA', logfunc=None):
        """checking the status of the solved issues"""
        par = super(SubmitDevtools, self)
        return par.check_issues_status(expected, logfunc=logger.warn)

    def check_branch(self, dummy, branch):
        """Check branch validity."""
        from hooks.devtools import repo_branch_checker
        return repo_branch_checker(branch) == OK


class AbstractSubmitSmeca(AbstractSubmit):

    """Minimal 'submit' command for some salome_meca tools (only push changes).
    """

    def __init__(self, ui, repo, *args, **opts):
        """Initialization"""
        super(AbstractSubmitSmeca, self).__init__(ui, repo, *args, **opts)
        self.refbranch = "edf/default"

    @classmethod
    def get_cfg(self, key, default=None):
        """Get a value of a parameter from the configuration."""
        return SMCFG.get(key, default)


class SubmitAsterStudy(AbstractSubmitSmeca):

    """Check AsterStudy source files at the parent revision
    and, in case of success, submit changesets for continuous integration.
    """
    command_automerge = ["./check_automerge_edf.sh"]
    command_lint = ["./check_lint.sh", "-p"]
    command_tests = [["./check_salome.sh", "-c"],
                     ["./check_docs.sh"],
                    ]

    @staticmethod
    def check_branch(dummy, branch):
        """Check branch validity."""
        from hooks.asterstudy import repo_branch_checker
        return repo_branch_checker(branch) == OK


class SubmitRTool(AbstractSubmitSmeca):

    """Submit command for some salome_meca RTOOL plugin."""

    @staticmethod
    def check_branch(dummy, branch):
        """Check branch validity."""
        from hooks.rtool import repo_branch_checker
        return repo_branch_checker(branch) == OK


class Integration(CandidateRevisionCheckAster):

    """Set of elementary commands to manage the requests for integration

    scenarii for integration:
    - for a single developer, merged revs to linear history
    - merge and linear two heads of submitted revisions
    """
    _interrupt_message = _("integration process interrupted because of errors")
    _cancel_message = _("integration cancelled")

    def __init__(self, ui, repo, *args, **opts):
        """Initialization"""
        super(Integration, self).__init__(ui, repo, *args, **opts)
        self._name = 'request'
        self._notify_on_error = False
        act = None
        for i in requestActions:
            if opts.get(i):
                act = i
        if act in ('start', 'differ', 'cancel', 'test', 'close', 'refuse'):
            if len(args) != 1:
                self._error(_("exactly one argument is expected"))
        elif act in ('check', 'startall', 'testall', 'closeall'):
            if len(args) > 0:
                self._error(_("no argument expected"))
        elif act in ('status',):
            pass
        else:
            self._error(_("unsupported action: '%s' (one of --status, --check... "
                          "is required)") % act)
        self._main = getattr(self, act)
        self.args = list(args)
        self._details = opts.get('message') or None
        if not self._details and act not in ('start', 'status', 'check'):
            ui.setconfig('ui', 'editor', self.get_cfg('editor2'))
            user = ui.username()
            self._details = ui.edit("", user)

    def check_branch(self, repo, branch):
        """Check branch validity."""
        return valid_branch(repo, branch)

    def _call_queue_func(self, func):
        """call ``func()`` on the revision passed in arguments"""
        rev = self.args[0]
        req = self._queue.get_by_node(rev)
        if req is None:
            logger.error(_("no request for this revision '%s'"), rev)
        else:
            iret = func(req.key)
            logger.update_from_status(iret)
            if logger.is_error():
                fmsg = tempfile.NamedTemporaryFile(prefix='msg.').name
                open(fmsg, 'wb').write(self._details)
                logger.warn(_("message written into: %s"), fmsg)
        self._check_error()

    def _loop_on_requests(self, status, func, repo=None, title=None):
        """Loop on requests in the given status and apply 'func' on each one
        If repo is not None, keep only requests in this repository."""
        self.init_queue()
        for req in self._queue:
            if req.status == status and repo in (None, req.repo) \
                    and req.branch == self.wrkbranch:
                self.args.append(req.node)
        while len(self.args):
            rev = self.args[0][:12]
            if title:
                logger.info(title, rev)
            func()
            self.args.pop(0)

    def start(self):
        """Mark a revision as started, its status is changed to RUN"""
        self.init_queue()
        func = partial(self._queue.start, details=self._details)
        self._call_queue_func(func)
        rev = self.args[0][:12]
        src = self.get_cfg('admin_push_uri')
        logger.title(_("pulling the revision %s from the remote server"), rev)
        iret = call([self.get_cfg('hg_bin'), 'pull', '-r', rev, src])
        if iret == 0:
            logger.info(_("pull ended successfully"))
        else:
            logger.warn(_("hg pull exits with code %s"), iret)
            logger.warn(_("You must pull the revision manually."))
        logger.info(
            _("next steps: hg rebase ..., hg up ..., hg request check"))

    def startall(self):
        """Start all requests in status PEND in the current repository"""
        self._loop_on_requests('PEND', self.start,
                               repo=self.reponame,
                               title=_("starting request for %s..."))

    def check(self):
        """Check of the source files,
        alert integrators on some specific warnings"""
        self.check_uncommitted()
        self.check_children()
        self.check_repositories_updates()
        self.build_changes_dict()
        self.check_message_descr()
        self.check_issues_status()
        self.init_waf_param()
        if self.reponame == 'src':
            self.check_waf_build()
            self.check_waf_install()
        self.init_report()
        self.check_source_files(askconfirm=False)
        self.report_on_new()
        # TODO filter report, highlight unrecommended warnings...
        logger.add_cr()
        self.status()
        logger.info(_("next step: hg request test or testall ..."))

    def differ(self):
        """Differ the integration process, the status of the request is
        reset to PEND"""
        self.init_queue()
        func = partial(self._queue.stop, details=self._details)
        self._call_queue_func(func)

    def cancel(self):
        """Cancel a request for integration, remove it from the queue"""
        self.init_queue()
        func = partial(self._queue.cancel, status=(
            'PEND', 'RUN'), details=self._details)
        self._call_queue_func(func)

    def close(self):
        """Revision has been integrated, remove its request from the queue"""
        self.init_queue()
        func = partial(self._queue.close, details=self._details)
        self._call_queue_func(func)

    def closeall(self):
        """Close all requests in status TEST"""
        self._loop_on_requests('TEST', self.close,
                               repo=self.reponame,
                               title=_("closing request for %s..."))

    def refuse(self):
        """Revision has been refused, remove its request from the queue"""
        self.init_queue()
        func = partial(self._queue.refuse, details=self._details)
        self._call_queue_func(func)

    def test(self):
        """Mark a revision to be tested"""
        self.init_queue()
        wrkrev = self.wrkrev
        strrev = str(wrkrev.rev()) + ':' + shortrev(self.wrkrev)
        logger.title(_("the revision {0} will become the new reference head "
                       "of {1}/{2}").format(strrev, self.reponame, self.wrkbranch))
        func = partial(self._queue.test, finalrev=wrkrev.hex(),
                       details=self._details)
        self._call_queue_func(func)
        logger.info(_("next step: run testcases"))

    def testall(self):
        """Mark all requests in status RUN to be tested"""
        self._loop_on_requests('RUN', self.test,
                               repo=self.reponame,
                               title=_("mark %s as TEST..."))

    def status(self):
        """Return the status of one or more requests"""
        self.init_queue()
        logger.title(_("status of the requests for integration:"))
        if self.args:
            for rev in self.args:
                req = self._queue.get_by_node(rev)
                if req is None:
                    logger.error(_("no request for this revision '%s'"), rev)
                else:
                    logger.info(repr(req))
        else:
            for req in self._queue:
                logger.info(repr(req))
            if len(self._queue) == 0:
                logger.info(_("no request"))

    def run(self):
        """Execute the action"""
        super(Integration, self).check_init()
        self.set_working_revision()
        self.set_parameters(need_changes=False)
        self._main()
        self._dump_log()


def _check_for_updates(reponame, path=None, update=False, **opts):
    """Check for updates"""
    logger.title(_("checking for updates for the repository '%s'"), reponame)
    if not path:
        path = ASCFG.get('devtools_root')
    retcode = check_for_updates(path, **opts)
    if retcode == 0:
        logger.error(_("there are available updates, please pull & update "
                       "in this repository before asking for the integration"))
        logger.error(_("you should run: %s"),
                     "cd %s && hg pull %s && cd %s"
                     % (path, '-u' * update, os.getcwdu()))
        if not update:
            logger.error(_("then you must update your development onto the "
                           "last 'reference' revision by using rebase or merge"))
    elif retcode != 1:
        logger.error(_("can not check for updates for the repository '{0}' "
                       "(options: {1})").format(reponame, opts))


def _check_description(descr):
    """Check a commit description
    The commit messages have already been checked by a dedicated hook but
    additional checkings can be done here."""
    linestar = [line for line in descr.splitlines() if line.startswith('***')]
    if linestar:
        logger.error('Message:\n' + descr)
        logger.error("A line starts with '***'. It seems to be a temporary "
                     "message of histedit. Please write a message that describe the "
                     "final revision.\n")


def _loglevel(ui):
    """Return the level to use in the logger object"""
    from aslint import logger
    lvl = logger.INFO
    if ui.configbool('ui', 'quiet'):
        lvl = logger.VERBOSE
    if ui.configbool('ui', 'debug'):
        lvl = logger.DEBUG
    return lvl
