#!/bin/bash

# --------------------------------------------------------------------
# Copyright (C) 1991 - 2017 - EDF R&D - www.code-aster.org
# This file is part of code_aster.
#
# code_aster is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# code_aster is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with code_aster.  If not, see <http://www.gnu.org/licenses/>.
# --------------------------------------------------------------------

TESTROOT=$(dirname $0)
LIBROOT=${TESTROOT}/../../lib
export PYTHONPATH=${PYTHONPATH}:${TESTROOT}

check_tests_usage()
{
    echo "Execute non-regression tests for the project."
    echo
    echo "Usage: $(basename $0) [options] [TEST ...]"
    echo
    echo "By default, executes tests in all modules. Optionally, module(s) to "
    echo "check can be explicitly specified to the script."
    echo
    echo "Examples:"
    echo "  $ $(basename $0) test_aslint.py"
    echo "  $ $(basename $0) test_hgaster.py:TestHgaster"
    echo "  $ $(basename $0) test_hgaster.py:TestHgaster.test02_pretxncommit"
    echo "  $ $(basename $0) test_aslint.py test_hgaster.py"
    echo
    echo "Options:"
    echo
    echo "  --help (-h)            Print this help information and exit."
    echo
    echo "  --parallel (-p)        Run tests in parallel."
    echo
    echo "  --coverage (-c)        Run with coverage."
    echo
    echo "  --report FILE          Report results in FILE (default: nosetests.xml)."
    echo
    echo "  --nocapture (-s)       Same as 'nosetests --nocapture' option."
    echo
    exit 0
}

check_tests_main()
{
    local parallel=0
    local coverage=0
    local nocapture=0
    local report=nosetests.xml

    local option
    while getopts ":-:hpcr:s" option $@
    do
        if [ "${option}" = "-" ]
        then
            case ${OPTARG} in
                help ) check_tests_usage ;;
                parallel ) parallel=1 ;;
                coverage ) coverage=1 ;;
                report ) report="${!OPTIND}" ; shift ;;
                nocapture ) nocapture=1 ;;
                * ) echo "Wrong option: --${OPTARG}" ; exit 1 ;;
            esac
        else
            case ${option} in
                h ) check_tests_usage ;;
                p ) parallel=1 ;;
                c ) coverage=1 ;;
                s ) nocapture=1 ;;
                ? ) echo "Wrong option" ; exit 1 ;;
        esac
        fi
    done
    shift $((OPTIND - 1))

    test "${coverage}" = "1" && parallel=0

    local options
    export ASLINT_WITHIN_TESTS=1

    options="${options} -c ${TESTROOT}/noserc"
    test "${parallel}" = "1" && options="${options} --processes=$(nproc)"
    test "${nocapture}" = "1" && options="${options} --nocapture"

    if [ "${coverage}" = "1" ]; then
        # Check that "pragma: no cover" directive usage decreases over time
        cover_debt_limit=0
        cover_debt=`find ${LIBROOT}/aslint -name "*.py" -exec grep -E "pragma:.+no.+cover" "{}" ";" | wc -l`
        if ((cover_debt > cover_debt_limit)); then
            echo "Cover debt (${cover_debt}) exceeds its limit (${cover_debt_limit})"
            return -1
        fi
        options="${options} --with-coverage --cover-erase"
        options="${options} --cover-min-percentage=64 --cover-xml --cover-xml-file=${report}"
    else
        options="${options} --with-xunit --xunit-file=${report}"
    fi

    local files
    if [ ${#} -gt 0 ]
    then
        local file path args
        for file in "${@}"
        do
            path=$(echo ${file} | awk -F: '{print $1}')
            path=$(find ${TESTROOT} -name "${path}")
            if [ "${path}" != "" ]
            then
                path=$(dirname $path)/${file}
                files="${files} ${path}"
            else
                echo "Error: can't find file ${file}"
                return 1
            fi
        done
    fi

    test "${files}" = "" && files="${TESTROOT}/aslint ${TESTROOT}/hgaster ${TESTROOT}/api"

    set -x

    nosetests ${options} ${files} || return 1
}

check_tests_main "$@"
exit $?
