# coding=utf-8

"""
Tests for hooks.
"""

import os
import unittest

from hamcrest import *


def test01_msg():
    """test message checking"""
    from hgaster.hooks.generic import OK, check_commit_msg, on_error
    ui = FakeUi()

    res = check_commit_msg(ui, "[#12345] valid message")
    assert_that(res, equal_to(OK))
    res = check_commit_msg(ui, "[#12345,#2222] valid message")
    assert_that(res, equal_to(OK))
    res = check_commit_msg(ui, "[#12345,2222] valid message")
    assert_that(res, equal_to(on_error()))
    res = check_commit_msg(ui, "#12345 invalid closure")
    assert_that(res, equal_to(on_error()))
    res = check_commit_msg(ui, "[#12] invalid message")
    assert_that(res, equal_to(on_error()))
    res = check_commit_msg(ui, "[#0000] tooshort")
    assert_that(res, equal_to(on_error()))
    res = check_commit_msg(ui, "[issue0000] nocomment")
    assert_that(res, equal_to(on_error()))
    res = check_commit_msg(ui, "[#9999] to NEW_COMMAND")
    assert_that(res, equal_to(on_error()))
    long_lines = os.linesep.join([
        "[#9999] to NEW_COMMAND", "", "bla " * 25
    ])
    res = check_commit_msg(ui, long_lines)
    assert_that(res, equal_to(on_error()))


def test02_branch_codeaster():
    from hgaster.hooks.generic import OK, NOOK
    from hgaster.hooks.codeaster import repo_branch_checker as checker

    assert_that(checker('default'), equal_to(OK))
    assert_that(checker('edf/mc'), equal_to(OK))
    assert_that(checker('asterxx/mc'), equal_to(OK))
    assert_that(checker('asterxxmc'), equal_to(NOOK))
    assert_that(checker('v13'), equal_to(OK))
    assert_that(checker('v18'), equal_to(OK))
    assert_that(checker('default_test'), equal_to(NOOK))
    assert_that(checker('issue8888'), equal_to(NOOK))


def test02_branch_check():
    from hgaster.hooks.generic import OK, NOOK, branch_checker

    checker = branch_checker(authorized='^edf/',
                             unauthorized=('^edf/default$', '^edf/build$'))
    assert_that(checker('edf/xxx'), equal_to(OK))
    assert_that(checker('edf/default'), equal_to(NOOK))
    assert_that(checker('edf/build'), equal_to(NOOK))
    assert_that(checker('edf/mybuild'), equal_to(OK))
    assert_that(checker('edf/build_test'), equal_to(OK))
    assert_that(checker('default'), equal_to(NOOK))


def test03_asterstudy():
    from hgaster.hooks.asterstudy import admin_or_contributors

    assert_that(admin_or_contributors("code_aster <code-aster@edf.fr>"),
                equal_to(True))
    assert_that(admin_or_contributors("vsr@opencascade.com"),
                equal_to(True))
    assert_that(admin_or_contributors("user@edf.fr"),
                equal_to(False))


class FakeUi(object):
    """Fake hg UI containing a minimal set of data for testing."""

    def __init__(self, user="user@edf.fr"):
        self.username = lambda: user
        self.warn = lambda msg: msg

class FakeCset(object):
    """Fake hg changeset containing a minimal set of data for testing."""

    def __init__(self, rev, user="user@edf.fr", branch="default", head=False):
        self.rev = lambda: rev
        self.hex = lambda: str(rev) * 10
        self.user = lambda: user
        self.branch = lambda: branch
        self.phase = lambda: 0
        self.description = lambda: "[#0123] changeset description"
        self._head = head
        self.extra = lambda: dict(close=not head)

class FakeRepo(object):
    """Fake hg repository containing a minimal set of data for testing."""

    def __init__(self, csets):
        self.dcsets = dict([i.rev(), i] for i in csets)
        self.dcsets[-1] = csets[-1]

    def __getitem__(self, key):
        return self.dcsets.get(key, self.dcsets[-1])

    def branchmap(self):
        dbr = dict()
        for rev, cset in self.dcsets.items():
            if rev < 0:
                continue
            dbr.setdefault(cset.branch(), [])
            if cset._head:
                dbr[cset.branch()].append(rev)
        return dbr

    def __len__(self):
        return len(self.dcsets)


def test04_commit_hook():
    from mercurial import error
    from hgaster.ext_utils import is_admin
    from hgaster.hooks.generic import (NOOK, OK, branch_user_commit_hook,
                                       branch_user_push_hook)

    def brch_chk(branch):
        if branch == "default":
            return OK
        return NOOK

    def priv_chk(user):
        return is_admin(user)

    pretxncommit = branch_user_commit_hook(brch_chk, priv_chk)
    pretxnchangegroup = branch_user_push_hook(brch_chk, priv_chk)

    ui = FakeUi()
    repo = FakeRepo([FakeCset(0), FakeCset(1)])
    assert_that(pretxncommit(ui, repo, hooktype='pretxncommit'), equal_to(OK))
    assert_that(pretxnchangegroup(ui, repo, 0), equal_to(OK))

    repo = FakeRepo([FakeCset(0, branch="test"), ])
    assert_that(calling(pretxncommit).with_args(ui, repo,
                                                hooktype='pretxncommit'),
                raises(error.Abort))
    assert_that(calling(pretxnchangegroup).with_args(ui, repo, 0),
                raises(error.Abort))

    ui = FakeUi(user='mergerobot')
    repo = FakeRepo([FakeCset(0, user='mergerobot')])
    assert_that(pretxncommit(ui, repo, hooktype='pretxncommit'), equal_to(OK))
    assert_that(pretxnchangegroup(ui, repo, 0), equal_to(OK))


def test05_devtools_hook():
    from mercurial import error
    from hgaster.hooks.generic import OK
    from hgaster.hooks.devtools import pretxncommit, pretxnchangegroup

    ui = FakeUi()
    repo = FakeRepo([FakeCset(0, branch='edf/user')])
    assert_that(pretxncommit(ui, repo, hooktype='pretxncommit'), equal_to(OK))
    assert_that(pretxnchangegroup(ui, repo, 0), equal_to(OK))

    repo = FakeRepo([FakeCset(0)])
    assert_that(calling(pretxncommit).with_args(ui, repo,
                                                hooktype='pretxncommit'),
                raises(error.Abort))
    assert_that(calling(pretxnchangegroup).with_args(ui, repo, 0),
                raises(error.Abort))

    ui = FakeUi(user='mergerobot')
    repo = FakeRepo([FakeCset(0, user='mergerobot')])
    assert_that(pretxncommit(ui, repo, hooktype='pretxncommit'), equal_to(OK))
    assert_that(pretxnchangegroup(ui, repo, 0), equal_to(OK))


def test05_asterstudy_hook():
    from mercurial import error
    from hgaster.hooks.generic import OK
    from hgaster.hooks.asterstudy import pretxncommit, pretxnchangegroup

    ui = FakeUi()
    repo = FakeRepo([FakeCset(0, branch='edf/user')])
    assert_that(pretxncommit(ui, repo, hooktype='pretxncommit'), equal_to(OK))
    assert_that(pretxnchangegroup(ui, repo, 0), equal_to(OK))

    repo = FakeRepo([FakeCset(0)])
    assert_that(calling(pretxncommit).with_args(ui, repo,
                                                hooktype='pretxncommit'),
                raises(error.Abort))
    assert_that(calling(pretxnchangegroup).with_args(ui, repo, 0),
                raises(error.Abort))

    ui = FakeUi(user='mergerobot')
    repo = FakeRepo([FakeCset(0, user='mergerobot')])
    assert_that(pretxncommit(ui, repo, hooktype='pretxncommit'), equal_to(OK))
    assert_that(pretxnchangegroup(ui, repo, 0), equal_to(OK))


def test_single_head():
    from hgaster.hooks.generic import OK, NOOK
    from hgaster.hooks import single_head_per_branch

    ui = FakeUi()
    repo = FakeRepo([
        FakeCset(0, branch='default'),
        FakeCset(1, branch='edf/user', head=True),
    ])
    assert_that(single_head_per_branch(ui, repo, repo[1],
                                       hooktype='pretxnchangegroup'),
                equal_to(OK))

    repo = FakeRepo([
        FakeCset(0, branch='default'),
        FakeCset(1, branch='edf/user', head=True),
        FakeCset(2, branch='edf/user', head=True),
    ])
    assert_that(single_head_per_branch(ui, repo, repo[1],
                                       hooktype='pretxnchangegroup'),
                equal_to(NOOK))


if __name__ == "__main__":
    import sys
    from testutils import get_test_suite
    RET = unittest.TextTestRunner(verbosity=2).run(get_test_suite(__name__))
    sys.exit(not RET.wasSuccessful())
